#include <assert.h>
#include <stdbool.h>
#include <errno.h>
#include <string.h>
#include <stdlib.h>
#include <platform.h>
#include <microhttpd.h>
#include <http_server.h>

#include <search.h>

#define	ONE_SERVER	1
#define	MULTIPLE_SERVERS	0

#if (ONE_SERVER && MULTIPLE_SERVERS) || (ONE_SERVER == 0 &&  MULTIPLE_SERVERS == 0)
#error "Forbidden configuration!"
#endif

/*
 * A tcp port on which the Http server will accept connections
 */
#define PORT 80

/*
 * A stack size for new tasks(threads)
 */
#define HTTP_SERVER_STACK_SIZE 2048

/*
 *  Memory allowed per connection.
 */
#define	MEMORY_CONNECTION_LIMIT		(8 * 1024)

/*
 * Name of HTTP server, header "Server:"
 */
#define HTTP_SERVER_NAME "WWW/1.0.0 (LIMA)"

/*
 * Maximum length of a username for  authentication.
 */
#define MAX_USERNAME_LENGTH (128 - 8)

/*
 * Maximum length of a password for authentication.
 */
#define MAX_PASSWORD_LENGTH (128 - 8)

/*
 * Maximum length of the URL.
 */
#define MAX_URL_LENGTH (128 - 8)

/*
 * A maximum number of supported URLs
 */
#define HTTP_SUPPORT_PAGES 64

/*
 * Default message for authentication realm.
 */
#define	AUTHENTICATION_REALM_MESSAGE	"Secure Area"


/*
 * Static variable to hold username for basic authentication.
 */
static char * username_ = NULL;

/*
 * Static variable to hold password for basic authentication.
 */
static char * password_ = NULL;


#if ONE_SERVER
/*
 * A hash table to search for URL handlers.
 */
static struct hsearch_data hash_table_;
#endif

/*
 * A page sent when the authentification process fails.
 */
#define	AUTH_FAIL_PAGE "<html><body>Authentification failed.</body></html>"

/*
 * A page sent when the URL handler is not found.
 */
#define	NOT_FOUND_PAGE "<html><body>Page not found!</body></html>"

/*
 * A struct to hold callbacks handling URLs.
 */
typedef struct URL_HANDLER
{
	URL_CALLBACK url_callback;
	REQUEST_COMPLETED_CALLBACK request_completed_callback;
} URL_HANDLER;


/* A function to check if given username and password match the stored */
static inline bool check_credentials(const char * username, const char * password)
{
	if ((NULL == username) || (NULL == password) || (NULL == username_) || (NULL == password_))
	{
		return false;
	}

	return (0 == strncmp(username, username_, MAX_USERNAME_LENGTH)) && (0 == strncmp(password, password_, MAX_PASSWORD_LENGTH));
}

/* A function called by MHD when something arrives */
static int AccessHandlerCallback(void *cls, struct MHD_Connection *connection, const char *url, const char *method, const char *version,
		const char *upload_data, size_t *upload_data_size, void **con_cls)
{
	(void)version;
	int hash_error;
	ENTRY ent, *entry;
	int ret = MHD_NO;
	char *user = NULL;
	char *pass = NULL;

	// Cast to get hash table with url handlers.
	struct hsearch_data * hash_table = (struct hsearch_data *) cls;

	// New connection?
	if (NULL == *con_cls)
	{
		// Response struct to add response.
		struct MHD_Response * response = NULL;

		// Data should never be greater than zero because of first call.
		assert((*upload_data_size) == 0);

		// Default behavior is FAIL.
		ret = MHD_NO;

		// get username and password
		user = MHD_basic_auth_get_username_password(connection, &pass);

		// check if it is valid
		if (false == check_credentials(user, pass))
		{
			// no, send denied reply
			response = MHD_create_response_from_buffer(strlen(AUTH_FAIL_PAGE), (void *) AUTH_FAIL_PAGE, MHD_RESPMEM_PERSISTENT);
			MHD_add_response_header(response, MHD_HTTP_HEADER_CONTENT_TYPE, "text/html" "; charset=iso-8859-2");
			// Set headers to always close connection
			MHD_add_response_header(response, MHD_HTTP_HEADER_CONNECTION, "close");
			ret = MHD_queue_basic_auth_fail_response(connection, AUTHENTICATION_REALM_MESSAGE, response);
			MHD_destroy_response(response);
		}
		else
		{
			// Valid, check if page exists.
			ent.key = (char *) url;

			// Get page from hash table
			hash_error = hsearch_r(ent, FIND, &entry, hash_table);

			// Check the return value of hsearch_r.
			if (0 == hash_error)
			{
				// url not found, send not found page
				response = MHD_create_response_from_buffer(strlen(NOT_FOUND_PAGE), (void *) NOT_FOUND_PAGE, MHD_RESPMEM_PERSISTENT);
				MHD_add_response_header(response, MHD_HTTP_HEADER_CONTENT_TYPE, "text/html" "; charset=iso-8859-2");
				// Set headers to always close connection
				MHD_add_response_header(response, MHD_HTTP_HEADER_CONNECTION, "close");
				// Not found response.
				ret = MHD_queue_response(connection, MHD_HTTP_NOT_FOUND, response);
				MHD_destroy_response(response);
			}
			else
			{
				// OK, set new HTTP_NEW_CONNECTION to continue.
				/*
				 * Thus, we will generate no response until the parameter is non-null—implying the callback was called before at least once.
				 * We do not need to share information between different calls of the callback, so we can set the parameter to any adress
				 * that is assured to be not null.
				 */
				*con_cls = HTTP_NEW_CONNECTION;
				ret = MHD_YES;
			}
		}

		// Add server name header
		MHD_add_response_header(connection->response, MHD_HTTP_HEADER_SERVER, HTTP_SERVER_NAME);

		// Dealocate because of MHD.
		free(user);
		free(pass);

		// Return value, close the connection if something is wrong.
		return ret;

	}

	// User's credentials and page exist were checked
	ent.key = (char *) url;
	hash_error = hsearch_r(ent, FIND, &entry, hash_table);

	// Check if the page exists.
	if (0 != hash_error)
	{
		// Match found - get url handler info struct
		URL_HANDLER * url_handler = (URL_HANDLER *) entry->data;

		assert(NULL != url_handler->url_callback);

		// Call the handler.
		ret = url_handler->url_callback(connection, method, upload_data, upload_data_size, con_cls);
	}
	else
	{
		// Very strange error, we have found it previously but not now?
		// Close the connection.
		return MHD_NO;
	}

	if (MHD_YES == ret)
	{
		// Add server name header
		MHD_add_response_header(connection->response, MHD_HTTP_HEADER_SERVER, HTTP_SERVER_NAME);
	}

	return ret;
}

/* A function called by MHD when connection ends */
static void RequestCompletedCallback(void *cls, struct MHD_Connection *connection, void **con_cls, enum MHD_RequestTerminationCode toe)
{
	ENTRY ent, *entry;
	int hash_error;

	// Cast to get hash table with url handlers.
	struct hsearch_data * hash_table = (struct hsearch_data *) cls;

	// Get url from connection struct.
	// DIRECT ACESS TO MHD_Connection struct !!!!!
	ent.key = (char *) connection->url;

	// Try to find RequestCompletedCallback for the url
	hash_error = hsearch_r(ent, FIND, &entry, hash_table);

	// Kein error and the callback defined?
	if ((0 != hash_error))
	{
		// Match found - get url handler info struct
		URL_HANDLER * url_handler = (URL_HANDLER *) entry->data;

		if (NULL != url_handler->request_completed_callback)
		{
			url_handler->request_completed_callback(connection, con_cls, toe);

		}
	}
}

/* Set credentials for the basic authentication. */
int http_server_set_credentials(const char * username, const char * password)
{
	int error = 0;
	size_t username_len = strlen(username); // C end string char
	size_t password_len = strlen(password); // C end string char

	if ((username_len >= MAX_USERNAME_LENGTH)|| (password_len >= MAX_PASSWORD_LENGTH))
	{
		error = -E2BIG;
	}
	else
	{
		// Dealocate
		free(username_);
		free(password_);

		// Alocate and copy
		username_ = strdup(username);
		password_ = strdup(password);

		// Check
		if((NULL == username_) || (NULL == password_))
		{
			// Dealocate
			free(username_);
			free(password_);
			username_ = password_ = NULL;
			// Set error
			error = -ENOMEM;
		}
	}
	return error;
}

/* Add url handler */
int http_server_add_url_handler(void * server, const char *url_prefix, URL_CALLBACK url_callback, REQUEST_COMPLETED_CALLBACK request_completed_callback)
{
	ENTRY ent, *ret;
	int hash_error = -ENOMEM;

#if ONE_SERVER
	(void) server;
	if (NULL == url_callback)
	{
		return -EINVAL;
	}
	// ONE_SERVER defined, use the static local hash_table.
	struct hsearch_data * hash_table = &hash_table_;
#elif MULTIPLE_SERVERS
	if((NULL == url_callback) || (NULL == server))
	{
		return -EINVAL;
	}
	struct hsearch_data * hash_table = (struct hsearch_data *)server;
#endif

	// Allocate structure to hold pointers.
	URL_HANDLER * url_entry = malloc(sizeof(URL_HANDLER));

	// Check malloc and len of url prefix
	if ((NULL != url_entry) && strlen(url_prefix) < MAX_URL_LENGTH)
	{
		// Fill hash structure.
		ent.key = strdup(url_prefix);
		ent.data = (void *) url_entry;

		if (NULL != ent.key)
		{
			// Fill url handler structure.
			url_entry->url_callback = url_callback;
			url_entry->request_completed_callback = request_completed_callback;

			// Try to store hash entry.
			hash_error = hsearch_r(ent, ENTER, &ret, hash_table);

			// Result?
			if (0 == hash_error)
			{
				// Error
				hash_error = -ENOMEM;
				// Deallocate url entry.
				free(ent.key);
				free(ent.data);
			}
			else
			{
				// OK.
				hash_error = 0;
			}
		}
	}
	else
	{
		// Deallocate url entry.
		free(url_entry);
	}

	// No free slot.
	return hash_error;
}

/* Start a http server.*/
void * http_server_start(MHD_AcceptPolicyCallback apc, void *apc_cls)
{
	int hash_error;
	struct MHD_Daemon *daemon;

	/* The memory necessary for the struct
	 hsearch_data object can be allocated dynamically. It must be initialized with zero before calling this
	 function.
	 */
#if ONE_SERVER
	struct hsearch_data * hash_table = &hash_table_;
#elif MULTIPLE_SERVERS
	struct hsearch_data * hash_table = calloc(1, sizeof(struct hsearch_data));
#endif
	// Check malloc result
	if (NULL != hash_table)
	{
		// Create hash table for pages.
		hash_error = hcreate_r(HTTP_SUPPORT_PAGES, hash_table);

		// Check hcreate_r ret val.
		if (0 != hash_error)
		{
			// Create HTTP server.
			daemon = MHD_start_daemon(MHD_USE_SELECT_INTERNALLY, PORT, apc, apc_cls, AccessHandlerCallback, (void *) hash_table, MHD_OPTION_NOTIFY_COMPLETED,
					RequestCompletedCallback, (void *) hash_table, MHD_OPTION_THREAD_STACK_SIZE, HTTP_SERVER_STACK_SIZE, MHD_OPTION_PER_IP_CONNECTION_LIMIT, 4,
					MHD_OPTION_CONNECTION_MEMORY_LIMIT, MEMORY_CONNECTION_LIMIT, MHD_OPTION_CONNECTION_LIMIT, 16, MHD_OPTION_CONNECTION_TIMEOUT, 16,
					MHD_OPTION_END);
			// Check the status of creating http daemon.
			if (NULL == daemon)
			{
				// Not enough memory or different error;
				// Destroy hash table.
				hdestroy_r(hash_table);
#if MULTIPLE_SERVERS
				// Deallocate hash_table pointer.
				free(hash_table);
#endif
				// Ret val.
				hash_table = NULL;

			}
		}
		else
		{
#if MULTIPLE_SERVERS
			// Deallocate hash_table pointer.
			free(hash_table);
#endif
			// Ret val.
			hash_table = NULL;
		}
	}

	// Return pointer to hash table, it will allow to add url handler.
	return (void *) hash_table;
}
