/*
 * http_server.h
 *
 *  Created on: 20.8.2012
 *      Author: velek
 */

#ifndef HTTP_SERVER_H_
#define HTTP_SERVER_H_

#include <stdbool.h>
#include <stdint.h>

#include "platform.h"
#include "microhttpd.h"
#include "internal.h"
#include "http_server_ssi.h"
#include "http_server_file.h"
#include "http_server_cgi.h"

/**
 * A value to determine new connection.
 */
#define		HTTP_NEW_CONNECTION	((void *)(~0))

/**
 * A helper macro to determine new connection in "if" statements.
 */
#define		HTTP_IS_NEW_CONNECTION(x)	((x) == (HTTP_NEW_CONNECTION))

/**
 * Signature of function called to handle requests.
 *
 * @param connection connection handle
 * @param method the HTTP method used ("GET", "PUT", etc.)
 * @param upload_data the data being uploaded (excluding HEADERS,
 *        for a POST that fits into memory and that is encoded
 *        with a supported encoding, the POST data will NOT be
 *        given in upload_data and is instead available as
 *        part of MHD_get_connection_values; very large POST
 *        data *will* be made available incrementally in
 *        upload_data)
 * @param upload_data_size set initially to the size of the
 *        upload_data provided; the method must update this
 *        value to the number of bytes NOT processed;
 * @param con_cls pointer that the callback can set to some
 *        address and that will be preserved by MHD for future
 *        calls for this request; since the access handler may
 *        be called many times (i.e., for a PUT/POST operation
 *        with plenty of upload data) this allows the application
 *        to easily associate some request-specific state.
 *        If necessary, this state can be cleaned up in the
 *        global "MHD_RequestCompleted" callback (which
 *        can be set with the MHD_OPTION_NOTIFY_COMPLETED).
 *        Initially, <tt>*con_cls</tt> will be NULL.
 * @return MHS_YES if the connection was handled successfully,
 *         MHS_NO if the socket must be closed due to a serious
 *         error while handling the request
 */
 typedef int
(*URL_CALLBACK)(struct MHD_Connection *connection, const char * method, const char *upload_data, size_t *upload_data_size,
		void **con_cls);

/**
 * Signature of the callback used by MHD to notify the
 * application about completed requests.
 *
 * @param connection connection handle
 * @param con_cls value as set by the last call to
 *        the MHD_AccessHandlerCallback
 * @param toe reason for request termination
 * @see MHD_OPTION_NOTIFY_COMPLETED
 */
typedef
void (*REQUEST_COMPLETED_CALLBACK)(struct MHD_Connection *connection, void **con_cls, enum MHD_RequestTerminationCode toe);

/**
 * Start a http server.
 *
 * @param apc callback to call to check which clients
 *        will be allowed to connect; you can pass NULL
 *        in which case connections from any IP will be
 *        accepted
 * @param apc_cls extra argument to apc
 * @return NULL on error, handle to daemon on success
 */
void * http_server_start(MHD_AcceptPolicyCallback apc, void *apc_cls);

/**
 * Add url handler.
 *
 * @param server	handle to daemon in MULTIPLE_SERVERS configuration or
 * 					NULL in ONE_SERVER.
 * @param url_prefix	URL to handle
 * @param url_callback	A callback function to handle URL
 * @param request_completed_callback	A callback function to notify the
 * 										application about completed requests
 *
 * @return Negative on error
 */
int http_server_add_url_handler(void * server, const char *url_prefix,
		URL_CALLBACK url_callback, REQUEST_COMPLETED_CALLBACK request_completed_callback);

#if 0
// Not defined on embedded systems
int http_server_stop(void * http_server_info);
#endif

/**
 * Set credentials for the basic authentication.
 *
 * @param username	Username
 * @param password	Password
 *
 * @return Negative on error
 */
int http_server_set_credentials(const char * username, const char * password);

#endif /* HTTP_SERVER_H_ */
