/*
 * http_server_cgi.c
 *
 *  Created on: 28.8.2012
 *      Author: velek
 */
/*
 * http_server_ssi.c
 *
 *  Created on: 27.8.2012
 *      Author: velek
 */

#include "platform.h"
#include <microhttpd.h>
#include <unistd.h>
#include <stdbool.h>
#include <string.h>
#include <stdint.h>

#include <http_server_cgi.h>

typedef struct
{
	size_t (*cgi_handler)(uint8_t ** buffer); /* Tag handler function */
	uint8_t * cgi_buffer; /* Number of bytes which should be copied to MHD buf. */
	size_t cgi_buffer_len; /* Number of bytes allocated by handler buffer. */
} WEB_FILE_CGI_DESCRIPTOR;

#define MHD_CREATE_RESPONSE_FROM_CALLBACK_BLOCK_SIZE	(512)

/* Text to display if CGI handler is not defined.*/
const char * const cgi_page_error = "<html><body>Page error!</body></html>";

/* Determine minimum of given arguments*/
static inline size_t min(size_t x, size_t y)
{
	// ret min
	return x < y ? x : y;
}

/* Tag handler function if not defined, it will copy the tag into buffer*/
static void cgi_callback_cleanup(void *cls)
{
	//cast the to file descriptor
	WEB_FILE_CGI_DESCRIPTOR * wfd = (WEB_FILE_CGI_DESCRIPTOR *) cls;

	if (NULL != wfd)
	{
		// Dealocate tag handler buffer if used. Must be NULL during initialization process.
		free(wfd->cgi_buffer);
		// Dealocate wfd structure.
		free(wfd);
	}
	return;
}

/* Tag handler function if not defined, it will copy the tag into buffer*/
static size_t cgi_default_handler(uint8_t ** buffer)
{
	// get len of tag

	// create buffer.
	*buffer = (uint8_t *) malloc(strlen(cgi_page_error));
	// check the creation
	if (NULL != *buffer)
	{
		// copy tag len to buffer.
		memcpy(*buffer, cgi_page_error, strlen(cgi_page_error));
		// return the amount of copied data.
		return strlen(cgi_page_error);
	}
	return 0;
}

static ssize_t cgi_content_generator(void *cls, uint64_t pos, char *buf,
		size_t buffer_len)
{

	// Return size, must not be zero.
	ssize_t size_to_return;

	//cast the to file descriptor
	WEB_FILE_CGI_DESCRIPTOR * wfd = (WEB_FILE_CGI_DESCRIPTOR *) cls;

	size_t buff_written_from_callback;
	uint8_t * tag_handler_buffer;

	// Check for first call.
	if (NULL == wfd->cgi_buffer)
	{
		// Not yet called, do it now.
		if (NULL != wfd->cgi_handler)
		{
			buff_written_from_callback = wfd->cgi_handler(&tag_handler_buffer);
		}
		else
		{
			// handler not define, use the default handler.
			buff_written_from_callback = cgi_default_handler(
					&tag_handler_buffer);
		}
		// Check for error.
		if (0 == buff_written_from_callback)
		{
#if 0 // Now handled in clean up function
			// Dealocate wfd structure.
			free(wfd);
#endif
			return MHD_CONTENT_READER_END_WITH_ERROR;
		}
		else
		{
			// Ok, setup variables.
			wfd->cgi_buffer = tag_handler_buffer;
			wfd->cgi_buffer_len = buff_written_from_callback;
		}
	}

	// No data left to parse, exit with end stream.
	if (pos >= wfd->cgi_buffer_len)
	{
#if 0 // Now handled in clean up function
		free(wfd->cgi_buffer);
		// Dealocate wfd structure.
		free(wfd);
#endif
		return MHD_CONTENT_READER_END_OF_STREAM;
	}
	// Determine the size which should be sent from buffer. Pos is always less than cgi_buffer_len, see the check above.
	size_t bytes_to_send_left = wfd->cgi_buffer_len - (size_t)pos;

	// Determine minimum of bytes which should be sent.
	size_t minimum = min(bytes_to_send_left, buffer_len);

	// copy to buffer.
	memcpy(buf, wfd->cgi_buffer + pos, minimum);

	// return the minimum of written data.
	size_to_return = minimum;

	// sanity check.
	if (size_to_return <= 0)
	{
#if 0 // Now handled in clean up function
		free(wfd->cgi_buffer);
		// Dealocate wfd structure.
		free(wfd);
#endif
		return MHD_CONTENT_READER_END_WITH_ERROR;
	}
	else
	{
		return size_to_return;
	}
}
#if 0
WEB_FILE_CGI_DESCRIPTOR * http_cgi_callback_setup(size_t (*cgi_handler)(uint8_t ** buffer))
{
	WEB_FILE_CGI_DESCRIPTOR * wfd = (WEB_FILE_CGI_DESCRIPTOR *) malloc(sizeof(WEB_FILE_CGI_DESCRIPTOR));
	if (NULL != wfd)
	{
		memset(wfd, 0, sizeof(WEB_FILE_CGI_DESCRIPTOR));
		wfd->cgi_buffer = NULL;
		wfd->cgi_handler = cgi_handler;
	}
	return wfd;
}
#endif
struct MHD_Response * http_server_cgi(size_t (*cgi_handler)(uint8_t ** buffer))
{

	struct MHD_Response *response = NULL;
	WEB_FILE_CGI_DESCRIPTOR * wfd = NULL;

	// Allocate info struct for ssi content
	wfd = (WEB_FILE_CGI_DESCRIPTOR *) malloc(sizeof(WEB_FILE_CGI_DESCRIPTOR));

	if (NULL != wfd)
	{
		// Init wfsd structure.
		memset(wfd, 0, sizeof(WEB_FILE_CGI_DESCRIPTOR));
		wfd->cgi_buffer = NULL;
		wfd->cgi_handler = cgi_handler;

		// Create response.
		response = MHD_create_response_from_callback(MHD_SIZE_UNKNOWN,
				MHD_CREATE_RESPONSE_FROM_CALLBACK_BLOCK_SIZE,
				cgi_content_generator, (void *) wfd, cgi_callback_cleanup);

		// If response is null, deallocate wfd
		if (NULL == response)
		{
			free(wfd);
		}

	}

	return response;
}

