/*
 * http_server_ssi.c
 *
 *  Created on: 27.8.2012
 *      Author: velek
 */

#include "platform.h"
#include <microhttpd.h>
#include <unistd.h>
#include <stdbool.h>
#include <string.h>
#include <stdint.h>

#include <http_server_ssi.h>

typedef enum TAG_CHECK_STATE
{
	TAG_NONE = 0, /* Not processing an SSI tag */
	TAG_LEADIN, /* Tag lead in "<!--#" being processed */
//	TAG_FOUND, /* Tag name being read, looking for lead-out start */
//	TAG_LEADOUT, /* Tag lead out "-->" being processed */
	TAG_SENDING /* Sending tag replacement string */
} TAG_CHECK_STATE;


typedef struct
{
	const uint8_t * file_content; /* Content of file. */
	size_t (*ssi_tag_handler)(const char * tag, uint8_t ** buffer); /* Tag handler function */
	const uint8_t * unprocessed; /* Pointer to the first unparsed byte in buf. */
	uint8_t * tag_handler_buffer; /* Number of bytes which should be copied to MHD buf. */
	size_t file_size;	/* File size. */
	size_t parse_left; /* Number of unparsed bytes in buf. */
	size_t data_to_copy; /* Number of bytes which should be copied to MHD buf during next call. */
	size_t tag_handler_buffer_len;	/* Number of bytes allocated by handler buffer. */
	size_t tag_handler_buffer_processed; /* Number of processed bytes of tag handler buffer. */

	enum TAG_CHECK_STATE tag_check_state; /* FSM for SSI Tags. */

} WEB_FILE_SSI_DESCRIPTOR;

/* Text to display if SSI page error is detected - Ussualy */
static const char * const ssi_page_error = "Page error!</body></html>";

#define MHD_CREATE_RESPONSE_FROM_CALLBACK_BLOCK_SIZE	(512)

/* The maximum length of the string comprising the tag name */
#define MAX_TAG_NAME_LEN 31

/* The SSI tag leading string */
#define LEN_TAG_LEAD_IN 5
static const char * const SSI_tag_lead_in = "<!--#";

/* The SSI tag ending string */
#define LEN_TAG_LEAD_OUT 3
static const char * const SSI_tag_lead_out = "-->";

/* Determine minimum of given arguments*/
static inline size_t min(size_t x, size_t y)
{
	// ret min
	return x < y ? x : y;
}

/* A function to generate error page during SSI parsing */
static ssize_t ssi_tag_error(const char * ssi_page_error, char *buf,
		size_t buffer_len)
{
	size_t minimum = min(strlen(ssi_page_error), buffer_len);
	// copy error message to buffer.
	memcpy(buf, ssi_page_error, minimum);
	// Set returned size.
	return (ssize_t) minimum;
}

/* Tag handler function if not defined, it will copy the tag into buffer*/
static size_t ssi_default_tag_handler(const char * tag, uint8_t ** buffer)
{
	// get len of tag
	size_t tag_len = strlen(tag);
	// create buffer.
	*buffer = (uint8_t *) malloc(tag_len);
	// check the creation
	if (NULL != *buffer)
	{
		// copy tag len to buffer.
		memcpy(*buffer, tag, tag_len);
		// return the amount of copied data.
		return tag_len;
	}
	return 0;
}

/* Tag handler function if not defined, it will copy the tag into buffer*/
static void ssi_callback_cleanup(void *cls)
{
	//cast the to file descriptor
	WEB_FILE_SSI_DESCRIPTOR * wfd = (WEB_FILE_SSI_DESCRIPTOR *) cls;

	if (NULL != wfd)
	{
		// Dealocate tag handler buffer if used. Must be NULL during initialization process.
		free(wfd->tag_handler_buffer);
		// Dealocate wfd structure.
		free(wfd);
	}
	return;
}

/* SSI file content generator */
static ssize_t ssi_content_generator(void *cls, uint64_t pos,
		char *buf, size_t buffer_len)
{
	(void) pos;

	// Return size, must not be zero.
	ssize_t size_to_return = MHD_CONTENT_READER_END_WITH_ERROR;

	//cast the to file descriptor
	WEB_FILE_SSI_DESCRIPTOR * wfd = (WEB_FILE_SSI_DESCRIPTOR *) cls;

	// No data left to parse, exit with end stream.
	if (0 == wfd->parse_left)
	{
#if 0 // Now handled in clean up function
		free(wfd->tag_handler_buffer);
		// Dealocate wfd structure.
		free(wfd);
#endif
		return MHD_CONTENT_READER_END_OF_STREAM;
	}

	// Check if wee need to process some data because of small buffer from the previous request.
	if (wfd->data_to_copy > 0)
	{
		// get minimum of remaining bytes to copy and len of given buffer
		size_t mininum = min(wfd->data_to_copy, buffer_len);
		// copy to buffer.
		memcpy(buf, wfd->unprocessed, mininum);
		// Decrease data to copy.
		wfd->data_to_copy -= mininum;
		// Decrease the parsed length.
		wfd->parse_left -= mininum;
		// Increase of processed bytes
		wfd->unprocessed += mininum;
		// return the minimum.
		return mininum;
	}

	switch (wfd->tag_check_state)
	{
	case TAG_NONE:
	{
		const uint8_t * tag_start;

		size_t data_to_copy;
		// check for starting SSI tag
		tag_start = memmem(wfd->unprocessed, wfd->parse_left, SSI_tag_lead_in,
				strlen(SSI_tag_lead_in));

		// did we found it?
		if (NULL == tag_start)
		{
			// NOT found
			data_to_copy = wfd->parse_left;

		} //if(NULL == wfd->tag_start)
		else
		{
			// yes, we did.
			wfd->tag_check_state = TAG_LEADIN;
			data_to_copy = (size_t)(tag_start - wfd->unprocessed);
		}

		size_t minimum = min(data_to_copy, buffer_len);
		// copy to buffer.
		memcpy(buf, wfd->unprocessed, minimum);
		// Decrease the parsed length.
		wfd->parse_left -= minimum;
		// Increase of processed bytes
		wfd->unprocessed += minimum;
		// setup size of copied data.
		size_to_return = minimum;

		// setup data which will be copied if the buffer is too short.
		wfd->data_to_copy = data_to_copy - minimum;

		break;
	}
	case TAG_LEADIN:
	{
		const uint8_t * tag_start;
		const uint8_t * tag_end;
		char tag_name[MAX_TAG_NAME_LEN + 1]; /* Tag name extracted */

		// check for closing SSI tag
		tag_end = memmem(wfd->unprocessed, wfd->parse_left, SSI_tag_lead_out,
				strlen(SSI_tag_lead_out));

		// did we found it or this is the last data in file or not tag handler found?
		if ((NULL == tag_end)
				|| ((size_t) ((tag_end + strlen(SSI_tag_lead_out))
						- wfd->unprocessed) >= wfd->parse_left))
		{
			// generate error into buffer.
			size_to_return = ssi_tag_error(ssi_page_error, buf, buffer_len);
			// Setup the descriptor to close processing.
			wfd->parse_left = 0;
		}
		else
		{
			// yes, we did. Process the tag.

			// Move pointer behind the tag.
			tag_start = wfd->unprocessed + strlen(SSI_tag_lead_in);

			// wfd->parse_left -= strlen(SSI_tag_lead_in);

			// determine length of tag.
			size_t tag_len = (size_t)(tag_end - tag_start);

			size_t buff_written_from_callback;
			uint8_t * tag_handler_buffer;

			if (tag_len <= MAX_TAG_NAME_LEN)
			{
				// copy the tag name.
				memcpy(tag_name, tag_start, tag_len);
				tag_name[tag_len] = '\0'; // add C string end character.

				// Call ssi tag handler to process data.
				if (NULL == wfd->ssi_tag_handler)
				{
					buff_written_from_callback = ssi_default_tag_handler(
							tag_name, &tag_handler_buffer);
				}
				else
				{
					buff_written_from_callback = wfd->ssi_tag_handler(tag_name,
							&tag_handler_buffer);
				}

				// Setup variables even if not processed...
				wfd->tag_handler_buffer = tag_handler_buffer;
				wfd->tag_handler_buffer_len = buff_written_from_callback;
				wfd->tag_handler_buffer_processed = 0;
			}

			if ((tag_len > MAX_TAG_NAME_LEN)
					|| (0 == buff_written_from_callback))
			{
				// Tag is too long or not found -> Error,
				// generate error into buffer.
				size_to_return = ssi_tag_error(ssi_page_error, buf, buffer_len);
				// Setup the descriptor to close processing.
				wfd->parse_left = 0;
				// exit case: statement
				break;
			}
			// Decrease the parsed length.
			wfd->parse_left -= (size_t)(tag_end + strlen(SSI_tag_lead_out)
					- wfd->unprocessed);
			// Increase of processed bytes
			wfd->unprocessed = tag_end + strlen(SSI_tag_lead_out);

			// Determine minimum of two buffers.
			size_t minimum = min(wfd->tag_handler_buffer_len, buffer_len);

			// copy to buffer.
			memcpy(buf, wfd->tag_handler_buffer, minimum);

			// Increment amount of data processed.
			wfd->tag_handler_buffer_processed += minimum;

			// check for remaining data.
			if (wfd->tag_handler_buffer_processed < wfd->tag_handler_buffer_len)
			{
				// We need to send data from buffer during next call.
				wfd->tag_check_state = TAG_SENDING;
			}
			else
			{
				// All data from tag processed, delete the buffer.
				free(wfd->tag_handler_buffer);
				wfd->tag_handler_buffer = NULL;
				wfd->tag_check_state = TAG_NONE;
			}
			// return the minimum of written data.
			size_to_return = minimum;
		}

		break;
	}
	case TAG_SENDING: // send data from tag buffer created by tag handler because the reply was too long to fit MHD buffer.
	{
		size_t minimum = min(
				wfd->tag_handler_buffer_len - wfd->tag_handler_buffer_processed,
				buffer_len);
		// copy to buffer.
		memcpy(buf, wfd->tag_handler_buffer + wfd->tag_handler_buffer_processed,
				minimum);

		// Increment amount of data processed.
		wfd->tag_handler_buffer_processed += minimum;

		// check if we should continue.
		if (wfd->tag_handler_buffer_len <= wfd->tag_handler_buffer_processed)
		{
			// All data from tag processed, delete the buffer.
			free(wfd->tag_handler_buffer);
			wfd->tag_handler_buffer = NULL;
			wfd->tag_check_state = TAG_NONE;
		}
		// return the minimum of written data.
		size_to_return = minimum;
		break;
	}
	} //switch

	// sanity check.
	if (size_to_return <= 0)
	{
#if 0 // Now handled in clean up function
		free(wfd->tag_handler_buffer);
		// Dealocate wfd structure.
		free(wfd);
#endif
		return MHD_CONTENT_READER_END_WITH_ERROR;
	}
	else
	{
		return size_to_return;
	}
}

#if 0
WEB_FILE_SSI_DESCRIPTOR * http_ssi_callback_setup(
		const uint8_t * ssi_file_content, size_t ssi_file_size,
		size_t (*ssi_tag_handler)(const char * tag, uint8_t ** buffer))
{
	WEB_FILE_SSI_DESCRIPTOR * wfd = (WEB_FILE_SSI_DESCRIPTOR *) malloc(
			sizeof(WEB_FILE_SSI_DESCRIPTOR));
	if (NULL != wfd)
	{
		memset(wfd, 0, sizeof(WEB_FILE_SSI_DESCRIPTOR));
		wfd->unprocessed = wfd->file_content = ssi_file_content;
		wfd->parse_left = wfd->file_size = ssi_file_size;
		wfd->tag_check_state = TAG_NONE;
		wfd->ssi_tag_handler = ssi_tag_handler;
		wfd->tag_handler_buffer = NULL;
	}
	return wfd;
}
#endif

struct MHD_Response * http_server_ssi(const uint8_t * ssi_file_content,
		size_t ssi_file_size,
		size_t (*ssi_tag_handler)(const char * tag, uint8_t ** buffer))
{

	struct MHD_Response *response = NULL;
	WEB_FILE_SSI_DESCRIPTOR * wfd = NULL;

	// Allocate info struct for ssi content
	wfd = (WEB_FILE_SSI_DESCRIPTOR *) malloc(
			sizeof(WEB_FILE_SSI_DESCRIPTOR));

	if (NULL != wfd)
	{
		// Init wfsd structure.
		memset(wfd, 0, sizeof(WEB_FILE_SSI_DESCRIPTOR));
		wfd->unprocessed = wfd->file_content = ssi_file_content;
		wfd->parse_left = wfd->file_size = ssi_file_size;
		wfd->tag_check_state = TAG_NONE;
		wfd->ssi_tag_handler = ssi_tag_handler;
		wfd->tag_handler_buffer = NULL;

		// Create response.
		response = MHD_create_response_from_callback(MHD_SIZE_UNKNOWN,
				MHD_CREATE_RESPONSE_FROM_CALLBACK_BLOCK_SIZE,
				ssi_content_generator, (void *) wfd,
				ssi_callback_cleanup);

		// If response is null, deallocate wfd
		if(NULL == response)
		{
			free(wfd);
		}

	}

	return response;
}

