/*
     This file is part of libmicrohttpd
     (C) 2008 Christian Grothoff (and other contributing authors)

     This library is free software; you can redistribute it and/or
     modify it under the terms of the GNU Lesser General Public
     License as published by the Free Software Foundation; either
     version 2.1 of the License, or (at your option) any later version.

     This library is distributed in the hope that it will be useful,
     but WITHOUT ANY WARRANTY; without even the implied warranty of
     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
     Lesser General Public License for more details.

     You should have received a copy of the GNU Lesser General Public
     License along with this library; if not, write to the Free Software
     Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
*/

/**
 * @file platform.h
 * @brief platform-specific includes for libmicrohttpd
 * @author Christian Grothoff
 *
 * This file is included by the libmicrohttpd code
 * before "microhttpd.h"; it provides the required
 * standard headers (which are platform-specific).<p>
 *
 * Note that this file depends on our configure.ac
 * build process and the generated config.h file.
 * Hence you cannot include it directly in applications
 * that use libmicrohttpd.
 */
#ifndef MHD_PLATFORM_H
#define MHD_PLATFORM_H

#include "MHD_config.h"

#define _XOPEN_SOURCE_EXTENDED  1
#if OS390
#define _OPEN_THREADS
#define _OPEN_SYS_SOCK_IPV6
#define _OPEN_MSGQ_EXT
#define _LP64
#endif

#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>
#include <string.h>
#include <unistd.h>
#include <stdarg.h>
#include <errno.h>
#include <fcntl.h>
#include <signal.h>
#include <string.h>
#if !defined(MINGW) && !defined(__SYMBIAN32__)
#include <search.h>
#endif
#include <stddef.h>
#undef HAVE_CONFIG_H
//#include <pthread.h>
#define HAVE_CONFIG_H 1

/* different OSes have fd_set in
   a broad range of header files;
   we just include most of them (if they
   are available) */


#ifdef OS_VXWORKS
#include <sockLib.h>
#include <netinet/in.h>
#include <stdarg.h>
#include <sys/mman.h>
#define RESTRICT __restrict__
#endif

#if HAVE_SYS_SELECT_H
#include <lwip/sockets.h>
#endif
#if HAVE_SYS_TYPES_H
#include <sys/types.h>
#endif
#if HAVE_SYS_TIME_H
#include <sys/time.h>
#endif
#if HAVE_SYS_STAT_H
#include <sys/stat.h>
#endif
#if HAVE_SYS_MSG_H
#include <sys/>
#endif
#if HAVE_SYS_MMAN_H
#include <sys/mman.h>
#endif
#if HAVE_NETDB_H
#include <lwip/netdb.h>
#endif
#if HAVE_NETINET_IN_H
#include <lwip/in.h>
#endif
#if HAVE_TIME_H
#include <time.h>
#endif
#if HAVE_SYS_SOCKET_H
#include <lwip/sockets.h>
#endif
#if HAVE_ARPA_INET_H
#include <lwip/inet.h>
#endif


#include <lima/system/rtos.h>

typedef int pthread_mutex_t;
typedef int pthread_t;
typedef size_t pthread_attr_t;
typedef int pthread_mutexattr_t;

static inline int pthread_mutex_lock(pthread_mutex_t * mutex)
{
	return mutex_lock(*mutex, RTOS_TIMEOUT_INFINITE);
}

static inline int pthread_mutex_unlock(pthread_mutex_t * mutex)
{
	return mutex_unlock(*mutex);
}

static inline int pthread_mutex_init(pthread_mutex_t *mutex, pthread_mutexattr_t * attr)
{
	(void) attr;
	int error = mutex_init(mutex, TASK_PRIORITY_NORMAL);
	if (error < 0)
		return -1;

	return 0;
}

static inline int pthread_mutex_destroy(pthread_mutex_t *mutex)
{
	mutex_destroy(*mutex);
	return 0;
}

#define	MHD_TASK_NAME	"Web server"
#define MHD_DEFAULT_PRIO	TASK_PRIORITY_LOW
#define MHD_DEFAULT_STACK_SIZE	1024


typedef struct MHD_TASK_PARAMS
{
	void *(*start_routine)(void*);
	void * arg;
}MHD_TASK_PARAMS;


static int rtos_mhd_thread_wrapper(int aMyTask, void *arg)
{
	(void) aMyTask;
	MHD_TASK_PARAMS * start_lwip_task = (MHD_TASK_PARAMS * )arg;
	start_lwip_task->start_routine(start_lwip_task->arg);
	return -EIO;
}


static inline int pthread_create(pthread_t * thread, const pthread_attr_t * attr, void *(*start_routine)(void*), void * arg)
{
	int stack_size;
	uint8_t * stack_area = NULL;

	if (NULL == attr)
	{
		stack_size = MHD_DEFAULT_STACK_SIZE;
	}
	else
	{
		stack_size = *attr;
	}

	MHD_TASK_PARAMS * start_lwip_task = (MHD_TASK_PARAMS *) malloc(sizeof(MHD_TASK_PARAMS));
	if (NULL == start_lwip_task)
	{
		return -1;
	}

	start_lwip_task->start_routine = start_routine;
	start_lwip_task->arg = arg;

	stack_area = (uint8_t *) malloc(stack_size);

	int task_id = task_create(rtos_mhd_thread_wrapper, (const char *) MHD_TASK_NAME, stack_size, stack_area, (void *) start_lwip_task, MHD_DEFAULT_PRIO);
	if (task_id < 0)
	{
		free(stack_area);
		return -1;
	}

	if (thread)
	{

		*thread = task_id;
	}

	return 0;
}

static inline int pthread_attr_init (void * attr)
{
	(void)attr;
	return 0;
}

static inline int pthread_attr_setstacksize(pthread_attr_t *attr, size_t stacksize)
{
	*attr = stacksize;
	return 0;
}

static inline int pthread_attr_destroy (pthread_attr_t *attr)
{
	(void)attr;
	return 0;
}

static inline int pthread_join(pthread_t thread, void **value_ptr)
{
	(void) thread;
	(void) value_ptr;
	return 0;
}
#if 0
#include <lwip/sys.h>
#include <arch/sys_arch.h>

typedef sys_mutex_t pthread_mutex_t;
typedef sys_thread_t pthread_t;
typedef size_t pthread_attr_t;
typedef int pthread_mutexattr_t;

static inline int pthread_mutex_lock(pthread_mutex_t * mutex)
{
	sys_mutex_lock(mutex);
	return 0;
}

static inline int pthread_mutex_unlock(pthread_mutex_t * mutex)
{
	sys_mutex_unlock(mutex);
	return 0;
}

static inline int pthread_mutex_init (pthread_mutex_t *mutex, pthread_mutexattr_t * attr)
{
	(void) attr;
	if(ERR_OK != sys_mutex_new(mutex))
		return -1;
	return 0;
}

static inline int pthread_mutex_destroy(pthread_mutex_t *mutex)
{
	sys_mutex_free(mutex);
	return 0;
}

#define	MHD_TASK_NAME	"Web server"
#define MHD_DEFAULT_PRIO	TCPIP_THREAD_PRIO
#define MHD_DEFAULT_STACK_SIZE	1024


typedef struct MHD_TASK_PARAMS
{
	void *(*start_routine)(void*);
	void * arg;
}MHD_TASK_PARAMS;


static int rtos_mhd_thread_wrapper(int aMyTask, void *aParameters)
{
	(void) aMyTask;

	MHD_TASK_PARAMS * start_lwip_task = (MHD_TASK_PARAMS * )arg;
	start_lwip_task->start_routine(start_lwip_task->arg);

	return -EIO;
}


static inline int pthread_create (pthread_t * thread,
        const pthread_attr_t * attr,
        void *(*start_routine)(void*), void * arg)
{
	int stack_size;
	uint8_t * stack_area = NULL;

	if(NULL == attr)
	{
		stack_size = MHD_DEFAULT_STACK_SIZE;
	}
	else
	{
		stack_size = *attr;
	}

	MHD_TASK_PARAMS * start_lwip_task = (MHD_TASK_PARAMS * )malloc(sizeof(MHD_TASK_PARAMS));
	if(NULL == start_lwip_task)
	{
		return -1;
	}

	start_lwip_task->start_routine = start_routine;
	start_lwip_task->arg = arg;

	stack_area = (uint8_t * )malloc(stack_size);

	CreatedTask = task_create(rtos_mhd_thread_wrapper, (const char *) MHD_TASK_NAME, stacksize, stack_area, (void *)start_lwip_task, (RTOS_TASK_PRIORITY) prio);
	if (CreatedTask < 0)
	{
		CreatedTask = SYS_THREAD_NULL;
	}

	*thread = sys_thread_new(MHD_TASK_NAME, mhd_lwip_thread_wrapper, start_lwip_task, stack_size, MHD_DEFAULT_PRIO);

	if(SYS_THREAD_NULL == *thread)
	{
		return -1;
	}

	return 0;
}

static inline int pthread_attr_init (void * attr)
{
	(void)attr;
	return 0;
}

static inline int pthread_attr_setstacksize(pthread_attr_t *attr, size_t stacksize)
{
	*attr = stacksize;
	return 0;
}

static inline int pthread_attr_destroy (pthread_attr_t *attr)
{
	(void)attr;
	return 0;
}

static inline int pthread_join(pthread_t thread, void **value_ptr)
{
	(void) thread;
	(void) value_ptr;
	return 0;
}
#endif

#include "plibc.h"

#endif
